(* Tart *)

open Unix;;
open Debug;;
open Problem;;
open Io;;
module G = Generic_io;;

(*** Specs *)
module Specs =
  struct
    open Opt
    let setopt x = Arg.String(fun y -> x := Some y);;
    let specs = Arg.align [
      "-trim",               Arg.Unit(fun () -> mode := Problem.Trim),    " find unsatisfiable variables";
      "-debug",              Arg.Set_int debug,                           " set debug level";
      "-check",              Arg.Set check,                               " check solution found";
      "-uninstallables",     Arg.Set uninstallables,                      " display only non-installables";
      "-dump-clauses",       setopt dump_clauses,                         "<file> dump clauses";
      "-cnf",                Arg.Set cnf,                                 " dump in industry-standard CNF format";
      "-human",              Arg.Set human,                               " load from human-readable file";
      "-machine",            Arg.Clear human,                              " load from machine-readable file";
      "-solver",             Arg.Set_string solver,                       "<solver> select solver module";
      "-print-variable-ids", Arg.Set print_variable_ids,                  " print variable ids in debugging output";
      "-benchmark",          Arg.Set benchmark,                           " do not print solutions";
      "-heuristics",         Arg.Set_string heuristics,                   "<heuristics> set solver-specific heuristics";
      "-log",                setopt log,                                  "<file> file for logging results";
      "-limit",              Arg.Set_int limit,                           "<num> maximum number of problems to solve";
      "-cost",               setopt cost,                                 "<file> file defining costs to minimize";
      "-min-cost",           Arg.Set_float min_cost,                      "<num> minimum cost";
      "-utility",            setopt utility,                              "<file> file defining utilities to maximize";
      "-max-count",          Arg.Set_int max_count,                       "<num> maximum number of nodes to explore";
      "-candidates",         Arg.String(fun u ->
                                let ul = Util.split_at ',' u in
                                candidates := !candidates @ ul),          "<variables> candidates to check";
      "-thin",               Arg.String(fun u ->
                                let ul = Util.split_at ',' u in
                                mode := Problem.Thin(List.map
                                  begin fun v ->
                                    try
                                      float_of_string v
                                    with
                                    | _ -> raise (Arg.Bad(sf "In -thin, bad float %S" v))
                                  end
                                  ul)),
                                                                          "<metric-list> size of media for thinning";
      "-max-solutions",      Arg.Set_int max_solutions,                   "<num> maximum number of solutions to consider for metric minization";
      "-nmb",                Arg.Set_int non_metrical_backtracking,       "<num> start metrical backtracking after this number of solutions";
    ]
  end
;;
(* ***)

let solvers_list = [
  "s1", Solver.solve;
  "s3", Solver3.solve
];;

let log_oc = lazy
  begin
    match !Opt.log with
    | None -> Pervasives.stdout
    | Some fn -> open_out fn
  end
;;

let total_time = ref 0.0;;
let max_time = ref 0.0;;
let num_problems = ref 0;;

exception Done;;

let solve fn =
  if !num_problems >= !Opt.limit then
    begin
      debug1 05 df "Limit of %d problems reached, terminating." !Opt.limit;
      raise Done
    end;
  let solve =
    try
      List.assoc !Opt.solver solvers_list
    with
    | Not_found ->
        Printf.eprintf "Unknown solver %S.  Known solvers: %s.\n"
          !Opt.solver
          (String.concat " " (List.map (fun (x,_) -> x) solvers_list));
        exit 1
  in
  debug1 05 df "Loading file %S" fn;
  let problem : problem =
    if !Opt.human then
      Human.load_from_file fn
    else
      G.load_from_file io_problem fn
  in
  debug0 05 df "Solving";
  let pt0 = times () in
  let solution = solve problem in
  let pt1 = times () in
  let t = pt1.tms_utime +. pt1.tms_stime -. pt0.tms_utime -. pt0.tms_stime in
  debug1 05 df "Total time %f." t;
  total_time := !total_time +. t;
  max_time := max !max_time t;
  incr num_problems;
  let oc = Lazy.force log_oc in
  begin
    match solution with
    | None ->
        Printf.fprintf oc "UNSAT %s %f\n" fn t
    | Some(Sequence soll) ->
        if not !Opt.benchmark then
          begin
            let i = ref 0 in
            debug1 10 df "Sequence of %d solutions" (List.length soll);
            List.iter
              begin fun sol ->
                debug2 10 df "Solution #%d has %d elements" !i (List.length sol.ssl_subset);
                List.iter
                  begin fun name ->
                    Printf.printf "%d %s\n" !i name
                  end
                  sol.ssl_subset;
                incr i
              end
              soll;
          end
    | Some(Set sol) ->
        if not !Opt.benchmark then
          List.iter
            begin fun name ->
              Printf.printf "%s\n" name
            end
            sol.ssl_subset;
        if !Opt.check then
          begin
            try
              Checker.check problem sol.ssl_subset;
              Printf.fprintf oc "** GOOD %s %f %f\n" fn sol.ssl_cost t
            with
            | Checker.Not_satisfied x ->
                Printf.fprintf oc "** BAD %s %f %f %s\n" fn sol.ssl_cost t x
          end
        else
          Printf.fprintf oc "** UNCHECKED %S %f %f\n" fn sol.ssl_cost t
  end;
  Printf.fprintf oc "%!"
;;

let rec over_directory f path =
  let dh = opendir path in
  try
    while true do
      match readdir dh with
      | "."|".." -> ()
      | entry ->
        let path' = Filename.concat path entry in
        dispatch f path'
    done
  with
  | End_of_file -> closedir dh
and dispatch f path =
  debug1 10 df "Dispatch over %S" path;
  let st = stat path in
  match st.st_kind with
  | S_DIR -> over_directory f path
  | _ -> f path
;;

let _ =
  Sys.catch_break true;
  begin
    try
      Arg.parse Specs.specs (dispatch solve) ((Filename.basename Sys.argv.(0)) ^ " [options]");
    with
    | Sys.Break -> ()
    | Done -> ()
  end;
  if !num_problems > 0 then
    begin
      debug4 01 df "Cumulated time %f for %d problems, %f per problem, max %f."
        !total_time
        !num_problems
        (!total_time /. (float !num_problems))
        !max_time
    end
;;
