(* Solver *)

open Debug;;
open Problem;;
open Clauses;;
open Unix;;

type reason = ISS.t;;

exception Fail;;
exception Next;;
exception Timeout;;

let fp = Printf.fprintf;;

(*** Conflicts *)
module Conflicts =
  struct
    type conflict_id = int;;
    type conflict = {
      id : int; (* ID number *)
      cardinality : int; (* Cardinality, for fast lookup. *)
      elements : IS.t; (* The actual elements. *)
    }
    type database = {
      mutable number : int; (* Total number of conflicts in database *)
      mutable index : IS.t IM.t; (* Map variables to conflict IDs *)
      mutable conflicts : conflict IM.t; (* Maps conflict ID to conflict *)
      mutable conflict_map : conflict_id ISM.t; (* Set of conflicts for fast membership *)
    }
    type watcher = {
      level : int;
      assignment : IS.t; (* Assigned variables *)
      levels : int IM.t;
      mutable matches_per_conflict : int IM.t; (* Maps conflict IDs to number of matching variables *)
      mutable last_known_number_of_conflicts : int; (* To know if this view is old or not. *)
    }

    exception Conflict of conflict_id;; (* A conflict has been detected. *)

    (*** number *)
    let number db = db.number;;
    (* ***)
    (*** dump database *)
    let dump_database db print_variable_set =
      debug1 20 df "Conflict database has %d conflicts:" db.number;
      IM.iter
        begin fun c_id cf ->
          debug3 30 df " c#%d %a" c_id print_variable_set cf.elements
        end
        db.conflicts
    ;;
    (* ***)
    (*** create_database *)
    let create_database () = 
      { number = 0;
        conflicts = IM.empty;
        conflict_map = ISM.empty;
        index = IM.empty }
    ;;
    (* ***)
    (*** register_conflict *)
    let register_conflict db ?(if_new=ignore) c =
      try
        ISM.find c db.conflict_map
      with
      | Not_found ->
          if_new c;
          let cf = {
            id = db.number;
            cardinality = IS.cardinal c;
            elements = c
          }
          in
          db.number <- db.number + 1;
          db.conflicts <- IM.add cf.id cf db.conflicts;
          db.conflict_map <- ISM.add c cf.id db.conflict_map;
          db.index <-
            IS.fold
              begin fun v index ->
                IM.add
                  v
                  (IS.add
                    cf.id
                    begin
                      try
                        IM.find v index
                      with
                      | Not_found -> IS.empty
                    end)
                  index
              end
              c
              db.index;
          cf.id
    ;;
    (* ***)
    (*** create_watcher *)
    let create_watcher db assignment =
      { matches_per_conflict =
          IM.map
            begin fun cf ->
              let vs = IS.inter cf.elements assignment in
              let m = IS.cardinal vs in
              if m = cf.cardinality then
                raise (Conflict cf.id)
              else
                m
            end
            db.conflicts;
        level = 0;
        levels = IM.empty;
        last_known_number_of_conflicts = db.number;
        assignment = assignment }
    ;;
    (* ***)
    (*** copy_and_assign *)
    let copy_and_assign db w v =
      let c_id_set =
        try
          IM.find v db.index
        with
        | Not_found ->
            IS.empty
      in
      let mpc =
        IS.fold
          begin fun c_id mpc ->
            let m =
              1 +
                 (try
                   IM.find c_id mpc
                 with
                 | Not_found -> 0)
            in
            let cf = IM.find c_id db.conflicts in
            if cf.cardinality = m then
              raise (Conflict cf.id)
            else
              IM.add 
                c_id
                m
                mpc
          end
          c_id_set
          w.matches_per_conflict
      in
      { matches_per_conflict = mpc;
        last_known_number_of_conflicts = w.last_known_number_of_conflicts;
        assignment = IS.add v w.assignment;
        level = w.level + 1;
        levels = IM.add v (w.level + 1) w.levels }
    ;;
    (* ***)
    (*** update *)
    let update db w =
      for c_id = w.last_known_number_of_conflicts to db.number - 1 do
        (* This watcher needs updating. *)
        let cf = IM.find c_id db.conflicts in
        let vs = IS.inter cf.elements w.assignment in
        let m = IS.cardinal vs in
        if m = cf.cardinality then
          raise (Conflict cf.id)
        else
          w.matches_per_conflict <- IM.add cf.id m w.matches_per_conflict
      done;
      w.last_known_number_of_conflicts <- db.number
    (* ***)
    (*** get_assignment *)
    let get_assignment db w = w.assignment;;
    (* ***)
    (*** get_conflict_elements *)
    let get_conflict_elements db c_id = (IM.find c_id db.conflicts).elements;;
    (* ***)
  end
;;
(* ***)
module C = Conflicts;;
(*module FQ = Persistent_queue;;*)
(*module FQ = Functional_queue;;*)
module FQ = Real_time_queue;;
(*** swap *)
let swap a i j =
  let x = a.(i) in
  a.(i) <- a.(j);
  a.(j) <- x
;;
(* ***)
(*** randomize *)
let randomize a =
  let m = Array.length a in
  for i = 1 to m - 1 do
    let j = i + Random.int (m - i) in
    swap a i j
  done
;;
(* ***)
(*** type env *)
type env = {
  depth : int;
  enqueued : IS.t; (* Normally this is redundant, since enqueued disjunctions will also be satisfied. *)
  watcher : C.watcher;
  queue : (int * int) FQ.t;
  cost : float;
};;
(* ***)
(*** type sol *)
type sol = {
  sol_variables : IS.t;
  sol_cost : float;
  sol_cardinal : int;
};;
(* ***)
(*** Solve *)
let solve problem =
  (*** Number variables *)
  let variables = Numbering.create () in
  let clauses = ref CS.empty in
  let string_of_clause (pos,neg) =
    sf "(%s,%s)"
      (String.concat " " (List.map (Numbering.name variables) (IS.elements pos)))
      (String.concat " " (List.map (Numbering.name variables) (IS.elements neg)))
  in
  List.iter
    begin fun (clause, reasons) ->
      let clause' =
        List.fold_left
          begin fun (pos,neg) (polarity, variable) ->
            let v = Numbering.register variables variable in
            if polarity then
              (IS.add v pos, neg)
            else
              (pos, IS.add v neg)
          end
          (IS.empty, IS.empty)
          clause
      in
      clauses := CS.add clause' !clauses
    end
    problem;
  (* ***)
  let clauses = !clauses in

  debug2 05 df "Total %d clauses and %d variables" (CS.cardinal clauses) (Numbering.number variables);

  let m = CS.cardinal clauses
  and n = Numbering.number variables
  in
  let root = n + 1 in

  (*** print_* *)
  (*** print_variable *)
  let print_variable oc v =
    if v = root then
      fp oc "root"
    else
      if !Opt.print_variable_ids then
        fp oc "%s#%d" (Numbering.name variables v) v
      else
        fp oc "%s" (Numbering.name variables v)
  in
  (* ***)
  (*** print_variable_set *)
  let print_variable_set oc vs =
    fp oc "{";
    IS.iter (fun v -> fp oc " %a" print_variable v) vs;
    fp oc " }";
  in
  (* ***)
  (*** print_variable_set_set *)
  let print_variable_set_set oc vss =
    fp oc "{";
    ISS.iter
      begin fun vs ->
        fp oc " %a" print_variable_set vs;
      end
      vss;
    fp oc " }";
  in
  (* ***)
  (*** print_clause *)
  let print_clause oc (pos, neg) =
    fp oc "(";
    IS.iter (fun v -> fp oc " %d(%s)" v (Numbering.name variables v)) pos;
    fp oc ",";
    IS.iter (fun v -> fp oc " %d(%s)" v (Numbering.name variables v)) neg;
    fp oc ")"
  in
  (* ***)
  (*** print_disjunction_set *)
  let print_disjunction_set oc ds =
    fp oc "{";
    IS.iter (fun id -> fp oc " #%d" id) ds;
    fp oc " }";
  in
  (* ***)
  (* ***)
  (*** metric_from_forcable *)
  let metric_from_forcable f =
    let a = Array.make n 0.0 in
    match Lazy.force f with
    | None -> a
    | Some m ->
        List.iter
          begin fun (vn, m) ->
            try
              let v_id = Numbering.find variables vn in
              a.(v_id) <- m
            with
            | Not_found -> ()
          end
          m;
        a
  in
  (* ***)
  (*** costs *)
  let costs = metric_from_forcable Globals.cost in
  let _ =
    for i = 0 to Array.length costs - 1 do
      costs.(i) <- max !Opt.min_cost costs.(i)
    done
  in
  (* ***)
  Dumper.dump variables clauses;
  (*** Compute variable conflict table and classify clauses *)
  let conflicts = Array.make n IS.empty in (* Maps every variable to its set of conflicting variables *)
  let roots = ref IS.empty in
  let unavailables = ref IS.empty in
  let forward = Array.make n ISS.empty in (* Maps variables to disjunctions *)
  let backward = Array.make n IS.empty in
  CS.iter
    begin fun ((pos, neg) as clause) ->
      (* Case I : Roots *)
      if IS.is_empty neg && IS.cardinal pos = 1 then
        roots := IS.union pos !roots
      else
        begin
          (* Case II : Conflicts *)
          if IS.is_empty pos && IS.cardinal neg = 2 then (* XXX encoding-specific *)
            begin
              match IS.elements neg with
              | [v1;v2] ->
                  conflicts.(v1) <- IS.add v2 conflicts.(v1);
                  conflicts.(v2) <- IS.add v1 conflicts.(v2)
              | _ -> assert false
            end
          else
            if IS.cardinal neg = 1 then
              if IS.is_empty pos then
                (* Case III : Unavailable *)
                unavailables := IS.union neg !unavailables
              else
                begin
                  (* Case IV : Dependencies *)
                  let v = IS.choose neg in
                  forward.(v) <- ISS.add pos forward.(v);
                  IS.iter
                    begin fun v' ->
                      backward.(v') <- IS.add v backward.(v')
                    end
                    pos
                end
            else
              Printf.printf "Unrecognized clause %a\n%!" print_clause clause
        end
    end
    clauses;
  (* ***)
  (*** Coalesce and number disjunctions. *)
  let disjunction_numbering = ref ISM.empty in
  let num_disjunctions = ref 0 in
  let forward' = Array.make n IS.empty in
  debug0 10 df "Numbering disjunctions.";
  for v = 0 to n - 1 do
    let ds = forward.(v) in (* Get the set of disjunctions required by the variable v *)
    ISS.iter
      begin fun d ->
        (* Iterate over disjunctions *)
        let (id, implicants) =
          try
            (* See if this disjunction has already been numbered *)
            ISM.find d !disjunction_numbering
          with
          | Not_found ->
              (* This is an unnumbered disjunction, number it *)
              let id = !num_disjunctions in
              incr num_disjunctions;
              let i = ref IS.empty in
              disjunction_numbering := ISM.add d (id, i) !disjunction_numbering;
              (id, i)
        in
        (* Register this variable as implying this disjunction *)
        implicants := IS.add v !implicants;
        (* Add this disjunction's ID to the set of disjunction IDs for this variable *)
        forward'.(v) <- IS.add id forward'.(v)
      end
      ds
  done;
  let num_disjunctions = !num_disjunctions in
  let disjunctions = Array.make num_disjunctions IS.empty in
  let implicants = Array.make num_disjunctions IS.empty in
  ISM.iter (fun d (id, vs_ref) ->
    disjunctions.(id) <- d;
    implicants.(id) <- !vs_ref) !disjunction_numbering;
  debug0 49 df "Disjunction numbers:";
  if_debug2 49
    Array.iteri (fun i d -> debug3 25 df "  #%d %a" i print_variable_set d) disjunctions;
  debug0 49 df "End of disjunction numbers.";
  (* ***)
  (*** Create and fill conflict database *)
  let cdb = C.create_database () in
  for v1 = 0 to n - 1 do
    let c = conflicts.(v1) in
    IS.iter
      begin fun v2 ->
        ignore (C.register_conflict cdb (IS.add v1 (IS.singleton v2)))
      end
      c
  done;
  IS.iter
    begin fun v ->
      ignore (C.register_conflict cdb (IS.singleton v))
    end
    !unavailables;
  C.dump_database cdb print_variable_set;
  (* ***)
  (*** Fill conflicts map. *)
  let forward = forward' in
  let conflicts' = ref IM.empty in
  for v1 = 0 to n - 1 do
    let c = conflicts.(v1) in
    conflicts' := IM.add v1 c !conflicts';
    if IS.is_empty c then
      debug2 40 df "Variable %d (%s) has no direct conflicts." v1 (Numbering.name variables v1)
  done;
  (* ***)
  (*** Variable to containing disjunctions map *)
  debug0 10 df "Filling variable to containing disjunctions map.";
  let variable_to_containing_disjunctions = Array.make n IS.empty in
  for id = 0 to num_disjunctions - 1 do
    let d = disjunctions.(id) in
    IS.iter
      begin fun v ->
        variable_to_containing_disjunctions.(v) <- IS.add id variable_to_containing_disjunctions.(v)
      end
      d
  done;
  (* ***)
  (*** Compute disjunction implication graph *)
  debug0 10 df "Computing disjunction implication graph.";
  let implication_graph_forward = Array.make num_disjunctions IS.empty
  and implication_graph_backward = Array.make num_disjunctions IS.empty
  in
  for id = 0 to num_disjunctions - 1 do
    let d1 = disjunctions.(id) in
    let v0 = IS.choose d1 in
    let containing_disjunctions =
      IS.fold
        begin fun v ds ->
          IS.inter ds variable_to_containing_disjunctions.(v)
        end
        d1
        variable_to_containing_disjunctions.(v0)
    in
    if IS.cardinal containing_disjunctions = 1 then
      debug1 41 df "Disjunction #%d has no implications." id
    else
      begin
        debug3 41 df "Disjunction #%d has the following implications: %a"
          id print_disjunction_set containing_disjunctions;
        implication_graph_forward.(id) <- containing_disjunctions
      end
  done;
  (* Build backward graph. *)
  for id1 = 0 to num_disjunctions - 1 do
    let ds = implication_graph_forward.(id1) in
    IS.iter
      begin fun id2 ->
        implication_graph_backward.(id2) <- IS.add id1 implication_graph_backward.(id2)
      end
      ds
  done;
  (* ***)
  (*** Initial environment. *)
  let env_conflicts = !conflicts' in
  let env0 = {
    depth = 0;
    enqueued = IS.empty;
    watcher = C.create_watcher cdb IS.empty;
    queue = FQ.empty;
    cost = 0.0
  }
  in
  let get_conflicts env v = IM.find v env_conflicts in
  let got_conflicts env =
    true
    (*IS.exists
      begin fun v ->
        not (IS.is_empty (IS.inter env.assignments (get_conflicts env v)))
      end
      env.assignments*) (* XXX *)
  in
  let get_assignment env = C.get_assignment cdb env.watcher in
  (*** enqueue_variable_dependencies *)
  let enqueue_variable_dependencies env v =
    (*let live_conflicts = IS.inter env.assignments (get_conflicts env v) in
    if not (IS.is_empty live_conflicts) then
      begin
        debug4 30 df "Cannot enqueue variable %a which has conflicts (%a)"
          print_variable v
          print_variable_set live_conflicts;
        raise Fail
      end
    else
      begin*)
        debug2 30 df "Enqueueing variable %a" print_variable v;
        let assignments = IS.add v (get_assignment env) in (* XXX *)
        let (queue, enqueued) =
          IS.fold (fun vs_id ((queue, enqueued) as q) ->
            (* See if this disjunction, or a stronger one has already been queued. *)
            (*if IS.mem vs_id enqueued then*)
            let ds = IS.inter enqueued implication_graph_backward.(vs_id) in
            if not (IS.is_empty ds) then
              begin
                debug1 35 df "Disjunction #%d or a stronger version already enqueued, ignoring" vs_id;
                q
              end
            else
              begin
                let vs = disjunctions.(vs_id) in
                if IS.is_empty (IS.inter assignments vs) then
                  (* Clause not satisfied yet. *)
                  begin
                    debug3 35 df "Adding disjunction #%d (%a)" vs_id print_variable_set vs;
                    (FQ.add (vs_id,v) queue, IS.add vs_id enqueued)
                  end
                else
                  (* Clause already satisfied *)
                  q
              end)
            forward.(v)
            (env.queue, env.enqueued)
        in
        { watcher = C.copy_and_assign cdb env.watcher v;
          cost = env.cost +. costs.(v);
          queue = queue;
          enqueued = enqueued;
          depth = env.depth + 1 }
      (*end*)
  in
  (* ***)

  debug0 25 df "Roots:";
  let env0 = 
    IS.fold
      begin fun v env ->
        debug2 25 df " %a" print_variable v;
        enqueue_variable_dependencies env v end
      !roots
      env0
  in
  (* ***)
  (*** in_degree *)
  let in_degree = Array.map IS.cardinal backward in
  let out_degree = Array.map
    begin fun ds ->
      IS.cardinal
        (IS.fold
          begin fun id set ->
            IS.union disjunctions.(id) set
          end
          ds
          IS.empty)
    end
    forward
  in
  if_debug2 45
    Array.iteri (fun v d -> debug3 45 df "  %a has in-degree %d" print_variable v d) in_degree;
  (* ***)
  (*** heuristically_order *)
  let heuristically_order =
    let sorter f =
      fun va ->
        Array.sort
          (fun v1 v2 ->
            let m1 = f v1 (*- out_degree.(v1)*)
            and m2 = f v2
            in
            compare m1 m2)
          va;
        va
    in
    match !Opt.heuristics with
    | "default"|"in-degree" -> sorter (fun v -> in_degree.(v))  
    | "out-degree" -> sorter (fun v -> out_degree.(v))  
    | "-in-degree" -> sorter (fun v -> - in_degree.(v))  
    | "-out-degree" -> sorter (fun v -> - out_degree.(v))  
    | "random" -> (fun va -> randomize va; va)
    | "identity" -> (fun va -> va)
    | "cost" -> sorter (fun v -> costs.(v))
    | "-cost" -> sorter (fun v -> -. costs.(v))
    | h -> invalid_arg (sf "Unknown heuristic %S" h)
  in                                                  
  (* ***)                                             
  (*** ordered_disjunctions *)                        
  let ordered_disjunctions =                          
    Array.map                                         
      (fun vs -> heuristically_order (Array.of_list (IS.elements vs)))
      disjunctions
  in
  (* ***)
  let first_solution = ref None in
  let best_solution = ref None in
  let best_cost = ref max_float in
  let solution_count = ref 0 in
  let improvements = ref 0 in
  let count = ref 0 in
  let installable = ref IS.empty in
  (*** reset_counters *)
  let reset_counters () =
    first_solution := None;
    best_solution := None;
    best_cost := max_float;
    solution_count := 0;
    improvements := 0;
    count := 0
  in
  (* ***)
  (*** The main ``breadth-first'' loop *)
  let rec bfs env =
    incr count;
    debug1 30 df "DEPTH %d" env.depth;
    debug2 39 df "  Assignments %a" print_variable_set (get_assignment env);
    if !Opt.max_count > 0 && !solution_count > 0 && !count > !Opt.max_count then raise Timeout;
    if !solution_count >= !Opt.non_metrical_backtracking && env.cost > !best_cost then raise Next;
    C.update cdb env.watcher;
    if FQ.is_empty env.queue then
      (*** Queue is empty, therefore all dependencies are satisfied.  Finish. *)
      begin
        debug0 30 df "Empty queue, finishing.";
        let cost = ref 0.0 in
        let subset = get_assignment env in
        let cost = env.cost
        and cardinal = IS.cardinal subset
        in
        let sol = { sol_variables = subset; sol_cost = cost; sol_cardinal = cardinal } in
        if !first_solution = None then first_solution := Some sol;
        if begin
              match !best_solution with
              | None -> true
              | Some sol' -> cost < sol'.sol_cost or
                             cost = sol'.sol_cost && IS.cardinal subset < sol'.sol_cardinal
            end
          then
            begin
              incr improvements;
              debug4 20 df "Solution #%d at count %d has %d elements and cost %f"
                !solution_count !count cardinal cost;
              best_cost := cost;
              best_solution := Some sol
            end
          else
            begin
              debug1 30 df "Solution #%d is not better." !solution_count;
            end;
        incr solution_count;
        if !solution_count >= !Opt.max_solutions then
          !best_solution
        else
          raise Next
      end
      (* ***)
    else
      (*** We still have disjunctive dependencies in the queue.
       *** Take the first one. *)
      begin
        let ((vs_id,ancestor),queue) = FQ.take env.queue in
        (* Someone needs one of the variables in vs to be true. *)
        let vs = disjunctions.(vs_id) in
        (* See if it is already satisfied *)
        if IS.is_empty (IS.inter vs (get_assignment env)) then
          (*** It is not already satisfied. *)
          begin
            (*** Heuristically order the element of the disjunction. *)
            let va = ordered_disjunctions.(vs_id) in
            (* ***)
            (*** debug *)
            debug5 31 df "  Considering disjunction #%d (%a) enqueued because of %a"
              vs_id
              begin fun oc va ->
                Array.iter
                  begin fun v ->
                    Printf.fprintf oc "%a=%d " print_variable v in_degree.(v)
                  end
                  va
              end
              va
              print_variable ancestor;
            (* ***)
            (*** Try the elements of the disjunction one by one. *)
            let rec attempt ?(hopeless=true) conflicts i =
              C.update cdb env.watcher;
              if i < Array.length va then
                (*** Let's attempt to set variable v to true *)
                (* See if v has any active conflicts. *)
                begin
                  let v = va.(i) in
                  try
                    debug2 32 df "  Setting %d (%s) to true." v (Numbering.name variables v);
                    let env = enqueue_variable_dependencies env v in
                    debug2 32 df "  Finished enqueuing dependencies for %d (%s)." v (Numbering.name variables v);
                    bfs env
                  with
                  | Next ->
                      if i = Array.length va - 1 then
                        raise Next
                      else
                        attempt ~hopeless:false conflicts (i + 1)
                  | Fail ->
                      debug3 32 df "FAILED choices for d%d=%a"
                        vs_id
                        print_variable_set vs;
                      attempt conflicts (i + 1) (* XXX *)
                  | C.Conflict c_id ->
                      if hopeless then
                        begin
                          (* This disjunction failed. *)
                          (* Find its implicants. *)
                          let ivs = implicants.(vs_id) in
                          debug5 32 df "CONFLICTING choices for d%d=%a whose implicants are %a"
                            vs_id
                            print_variable_set vs
                            print_variable_set ivs;
                          (* Propagate conflicts here *)

                          attempt (IIS.add (v,c_id) conflicts) (i + 1)
                        end
                      else
                        begin
                          (* This disjunction leads to a solution, but we are optimizing.
                           * Do not propagate conflicts. *)
                          attempt ~hopeless conflicts (i + 1)
                        end
                end
                (* ***)
              else
                begin
                  (*** All attemps have failed.  This disjunction cannot be presently satisfied. *)
                  if hopeless then
                    begin
                      debug1 30 df "  All attempts for disjunction #%d have failed." vs_id;
                      let cs = 
                        IIS.fold
                          begin fun (v,c_id) cs ->
                            let cs0 = C.get_conflict_elements cdb c_id in
                            debug5 41 df "    variable %a conflict #%d %a" print_variable v c_id print_variable_set cs0;
                            IS.union (IS.remove v cs0) cs
                          end
                          conflicts
                          IS.empty
                      in
                      (* XXX: here we arbitrarily choose one conflict of interest. *)
                      let ivs = implicants.(vs_id) in
                      IS.iter
                        begin fun v ->
                          ignore
                            (C.register_conflict
                              ~if_new:
                                begin fun cs ->
                                  debug2 21 df "  Learned conflict %a" print_variable_set cs;
                                end
                              cdb (IS.add v cs))
                        end
                        (IS.inter (get_assignment env) ivs);

                      let c_id = C.register_conflict cdb (IS.add ancestor cs) in
                      raise (C.Conflict c_id)
                      (* ***)
                    end
                  else
                    raise Next
                end
            in
            (* ***)
            attempt IIS.empty 0
          end
          (* ***)
        else
          begin
            (*** Disjunction already satisfied, continue. *)
            debug2 35 df "  Disjunction %a already satisfied" print_variable_set vs;
            bfs { env with queue = queue }
            (* ***)
          end
      end
      (* ***)
  in
  (* ***)
  let pt0 = times () in
  debug2 10 df "Starting search at user time %f system time %f" pt0.tms_utime pt0.tms_stime;
  (*** dopt *)
  let dopt = if_debug1 10
    begin fun pt1 ->
      let user   = pt1.tms_utime -. pt0.tms_utime
      and system = pt1.tms_stime -. pt0.tms_stime
      in
      debug2 10 df "User %f system %f" user system
    end
  in
  (* ***)
  (*** set_to_name_list *)
  let set_to_name_list set = IS.fold (fun v result -> (Numbering.name variables v) :: result) set [] in
  (* ***)
  (*** convert *)
  let convert = function
    | Some sol ->
        begin
          match !first_solution with
          | None -> 
              ()
          | Some sol' ->
              debug5 10 df "Total %d improvements in %d solutions (%05.2f%% from %f to %f)"
                !improvements
                !solution_count
                (100.0 *. ((sol'.sol_cost -. sol.sol_cost) /. sol'.sol_cost))
                sol'.sol_cost
                sol.sol_cost;
        end;
        Some(Set{ ssl_subset = set_to_name_list sol.sol_variables;
                  ssl_cost = sol.sol_cost })
    | None -> None
  in
  (* ***)
  (*** install *)
  let install () =
    let env =
      List.fold_left
        begin fun env vn ->
          let v = Numbering.find variables vn in
          enqueue_variable_dependencies env v
        end
        env0
        !Opt.candidates
    in
    try
      let y = bfs env in
      convert y
    with
    | Next|Timeout -> convert !best_solution
  in
  (* ***)
  (*** closure_size *)
  let closure_size v =
    let visited = Array.make n false in
    let rec dfs set v =
      visited.(v) <- true;
      let set = set + 1 in (* IS.add v set in *)
      IS.fold
        begin fun d_id set ->
          let ds = disjunctions.(d_id) in
          IS.fold
            begin fun v' set ->
              if not visited.(v') then
                dfs set v'
              else
                set
            end
            ds
            set
        end
        forward'.(v)
        set
    in
    dfs 0 (* IS.empty *) v
  in
  (* ***)
  (*** trim *)
  let trim () =
    let utility = in_degree in
    let unknown = ref IIS.empty in
    let uninstallable = ref IS.empty in
    begin
      match !Opt.candidates with
      | [] ->
        for v = 0 to n - 1 do
          unknown := IIS.add (in_degree.(v),v) !unknown;
        done
      | l -> List.iter (fun vn ->
            try
              let v = Numbering.find variables vn in
              unknown := IIS.add (in_degree.(v),v) !unknown
            with
            | Not_found ->
                debug1 05 df "Unknown variable %S" vn) l
    end;
    debug1 10 df "Will trim a set of %d variables" n;
    while not (IIS.is_empty !unknown) do
      let ((_,v) as v_e) = IIS.min_elt !unknown in
      unknown := IIS.remove v_e !unknown;
      debug3 15 df "Checking installability of %a (%d left)" print_variable v (IIS.cardinal !unknown);
      match
        try
          let env = enqueue_variable_dependencies env0 v in
          reset_counters ();
          bfs env
        with
        | Next ->
            !best_solution
        | C.Conflict c_id ->
            let cs = C.get_conflict_elements cdb c_id in
            debug4 11 df "Conflict for %a : %a" print_variable v print_variable_set cs;
            None
        | x ->
            debug1 20 df "Exception %s" (Printexc.to_string x);
            None
      with
      | None ->
          uninstallable := IS.add v !uninstallable;
          debug2 20 df "Uninstallable: %a" print_variable v
      | Some sol ->
          IS.iter
            begin fun v' ->
              if not (IS.mem v' !installable) then
                begin
                  unknown := IIS.remove (in_degree.(v'),v') !unknown;
                  debug2 10 df "Installable: %a" print_variable v';
                  installable := IS.add v' !installable
                end
            end
            sol.sol_variables
    done;
    let x = if !Opt.uninstallables then !uninstallable else !installable in
    convert(Some { sol_variables = x;
                   sol_cost = 0.0;
                   sol_cardinal = IS.cardinal x })
  in
  (* ***)
  (*** cost_of_set *)
  let cost_of_set set = IS.fold (fun v size -> size +. costs.(v)) set 0.0 in
  (* ***)
  (*** thin *)
  let module FIS = Set.Make(struct type t = float * int let compare = compare end) in
  let thin media =
    debug0 10 df "Thinning";
    let candidates = ref FIS.empty in
    let candidates_by_size = ref FIS.empty in
    let util = metric_from_forcable Globals.utility in
    for v = 0 to n - 1 do
      candidates := FIS.add (util.(v), v) !candidates;
      candidates_by_size := FIS.add (costs.(v), v) !candidates_by_size
    done;
    let uninstallable = ref IS.empty in
    let rec loop set cost candidates candidates_by_size limit =
      debug1 20 df "Cost is %f" cost;
      if FIS.is_empty candidates then
        begin
          debug1 10 df "Finished, no candidates remain; cost is %f" cost;
          set
        end
      else
        begin
          let (p,v) = FIS.max_elt candidates in
          let candidates = FIS.remove (util.(v), v) candidates in
          let candidates_by_size = FIS.remove (costs.(v), v) candidates_by_size in
          if cost +. costs.(v) > limit then
            begin
              let (_,v') = FIS.min_elt candidates_by_size in
              if cost +. costs.(v') > limit then
                begin
                  debug3 11 df "Cost is %f, smallest candidate is %f, limit is %f:  No more plausible candidates"
                    cost
                    costs.(v')
                    limit;
                  set
                end
              else
                begin
                  debug4 20 df "Can't use candidate %a with utility %f; cost is %f" print_variable v p cost;
                  loop set cost candidates candidates_by_size limit
                end
            end
          else
            begin
              debug4 20 df "Trying candidate %a with utility %f; cost is %f" print_variable v p cost;
              match
                try
                  let env = enqueue_variable_dependencies env0 v in
                  reset_counters ();
                  bfs env
                with
                | Next|Timeout ->
                    begin
                      if !best_solution = None then debug0 11 df "Warning: no solution";
                      !best_solution
                    end
                | Fail ->
                    debug2 19 df "Failure for %a" print_variable v;
                    None
                | Conflicts.Conflict c_id ->
                    let cs = C.get_conflict_elements cdb c_id in
                    debug4 19 df "Conflict for %a : %a" print_variable v print_variable_set cs;
                    None
                | x ->
                    debug3 10 df "While trying %a, exception %s" print_variable v (Printexc.to_string x);
                    raise x
              with
              | None ->
                  debug2 10 df "Candidate %a is not installable" print_variable v;
                  uninstallable := IS.add v !uninstallable;
                  loop set cost candidates candidates_by_size limit
              | Some sol ->
                  let set' = sol.sol_variables in
                  let set' = IS.union set set' in
                  let cost' = cost_of_set set' in
                  let (set'', cost'') =
                    if cost' > limit then
                      begin
                        debug2 20 df "Cost %f exceeds limit %f" cost limit;
                        (set, cost)
                      end
                    else
                      (set', cost')
                  in
                  try
                    loop set'' cost'' candidates candidates_by_size limit
                  with
                  | Sys.Break as x -> raise x
                  | _ -> set

            end
        end
    in
    let (solution,_,_) =
      List.fold_left
        begin fun (result,set,cumulated_media) medium ->
          debug1 10 df "Computing next medium of size %f" medium;
          let (candidates, candidates_by_size) =
            IS.fold
             begin fun v (candidates, candidates_by_size) ->
               (FIS.remove (util.(v),v) candidates,
                FIS.remove (costs.(v),v) candidates_by_size)
             end
             (IS.union set !uninstallable)
             (!candidates, !candidates_by_size)
          in
          let set' =
            loop
              set
              cumulated_media
              candidates
              candidates_by_size
              (cumulated_media +. medium)
          in
          let total_utility =
            IS.fold
              begin fun v utility ->
                util.(v) +. utility
              end
              set'
              0.0
          in
          debug1 10 df "Total utility is %f" total_utility;
          let mset = IS.diff set' set in
          (({ ssl_subset = set_to_name_list mset;
              ssl_cost = cost_of_set set' } :: result),
           set',
           cumulated_media +. medium)
        end
        ([], IS.empty, 0.0)
        media
    in
    Some(Sequence (List.rev solution))
  in
  (* ***)
  (*** process *)
  let process () =
    match !Opt.mode with
    | Problem.Install -> install ()
    | Problem.Trim -> trim ()
    | Problem.Thin media -> thin media
    (*| _ -> invalid_arg "Mode not implemented."*)
  in
  (* ***)
  (*** Main wrapper *)
  try
    let y = process () in
    let pt1 = times () in
    dopt pt1;
    y
  with
  | x ->
    let pt1 = times () in
    dopt pt1;
    raise x
  (* ***)
;;
(* ***)
(*** Small wrapper for uncaught backtracks. *)
let solve problem =
  try
    solve problem
  with
  | Fail -> None
  | C.Conflict _ -> None
;;
(* ***)
